from __future__ import annotations

import numpy as np
import pandas as pd


SECONDS_PER_YEAR = 31_557_600


@pd.api.extensions.register_series_accessor("pt")
class SeriesAccessor:
    def __init__(self, pandas_obj: pd.Series):
        self._obj = pandas_obj

    def downcast(self):
        s = self._obj
        if s.dtype == "float64":
            all_int = np.all(np.mod(s.to_numpy(), 1) == 0)
            dc_type = "integer" if all_int else "float"
        elif s.dtype == "int64":
            dc_type = "integer"
        else:
            return self._obj
        s = pd.to_numeric(s, downcast=dc_type)
        return self._obj

    def tolerance_bands(self, window, pct):
        series = self._obj
        rolling = series.rolling(window=window, center=True, min_periods=1)
        u_band = np.maximum(rolling.max(), series * (pct + 1))
        l_band = np.minimum(rolling.min(), series * (1 - pct))
        result = pd.concat([series, l_band, u_band], axis=1)
        cols = [f"{series.name}_{x}" for x in ["real", "lower", "upper"]]
        result.columns = cols
        return result

    def float_year_to_datetime(self):
        year = self._obj.astype(int)
        secs = (self._obj - year) * SECONDS_PER_YEAR
        return pd.to_datetime(year, format="%Y") + pd.to_timedelta(secs, unit="s")

    def cut(
        self,
        bins,
        right=True,
        labels=None,
        retbins=False,
        precision=3,
        include_lowest=False,
        duplicates="raise",
        ordered=True,
    ):
        return pd.cut(
            self._obj,
            bins=bins,
            right=right,
            labels=labels,
            retbins=retbins,
            precision=precision,
            include_lowest=include_lowest,
            duplicates=duplicates,
            ordered=ordered,
        )


if __name__ == "__main__":
    df = pd.DataFrame(dict(a=[1, 2, 3, 4, 5]))
    print(df["a"].pt.tolerance_bands(5, 0.01))



from __future__ import annotations

from collections.abc import Hashable
import io
import logging

import numpy as np
import pandas as pd

from pandastools.utils import dataimport, helpers


logger = logging.getLogger(__name__)


@pd.api.extensions.register_dataframe_accessor("pt")
class DataFrameAccessor:
    def __init__(self, parent):
        self._obj = parent

    def iat(self, start=None, stop=None, step=1):
        return self._obj.iloc[start:stop:step]

    def get_info(self, show_counts: bool = True) -> str:
        """Get a concise summary of a DataFrame.

        This method prints information about a DataFrame including
        the index dtype and column dtypes, non-null values and memory usage.

        Parameters
        ----------
        show_counts : bool, optional
            Whether to show the non-null counts.

        Returns
        -------
        str
            The summary of a DataFrame.
        """
        buf = io.StringIO()
        self._obj.info(buf=buf, show_counts=show_counts)
        return buf.getvalue()

    def uniquify_columns(self) -> pd.DataFrame:
        """Modify column names of a DataFrame to be unique.

        In case a DataFrame contains several equal column names, rename all columns
        except the first one by appending _{count}.

        Returns
        -------
        pd.DataFrame
            A DataFrame with unique column names
        """
        diff = len(self._obj.columns) - len(set(self._obj.columns))  # type: ignore
        if diff == 0:
            return self._obj
        seen = set()
        new = list()
        for item in self._obj.columns:  # type: ignore
            fudge = 1
            newitem = str(item)
            while newitem in seen:
                fudge += 1
                newitem = f"{item}_{fudge}"
            seen.add(newitem)
            new.append(newitem)
        df = self._obj.copy()
        df.columns = new
        return df

    def convert_dtypes(self, old_type, new_dtype):
        cols = self._obj.select_dtypes([old_type]).columns
        df = self._obj.copy()
        df[cols] = df.select_dtypes([old_type]).apply(lambda x: x.astype(new_dtype))
        return df

    def split(self, thresh: float, colname: str, extra_rows: int = 0) -> pd.DataFrame:
        """Append columns with split information based on supplied criteria.

        Parameters
        ----------
        thresh : float
            Threshold to use for splitting.
        colname : str
            Column used to detect splits.
        extra_rows : int, optional
            Amount of rows to append in both directions for each split section.

        Returns
        -------
        pd.DataFrame
            The dataframe including new columns containing the split information.
        """
        df = self._obj.drop("secs", errors="ignore", axis=1)
        array = np.full((len(df.index),), np.nan)
        df["process_num"] = pd.Categorical(array)
        df = dataimport.add_transition_info(
            ds=df, colname=colname, threshold=thresh, extra_rows=extra_rows
        )
        return df

    def index_to_secs(self) -> pd.DataFrame:
        """Convert the DateTimeIndex to seconds, starting with 0.

        Returns
        -------
        pd.DataFrame
            DataFrame containing an IntegerIndex
        """
        if self._obj.empty:
            logger.debug("index_to_secs failed. Dataframe empty")
            return self._obj
        elif isinstance(self._obj.index, pd.DatetimeIndex):
            secs = self._obj.index.astype(int) / 1_000_000_000
            df = self._obj.assign(secs=secs - secs[0]).set_index("secs", drop=True)
            return df
        else:
            logger.debug("index_to_secs failed. No DateTimeIndex")
            return self._obj

    def cleanup(self):
        df = self._obj.infer_objects()
        cols = df.select_dtypes(["object"]).columns
        df[cols] = df.select_dtypes(["object"]).apply(lambda x: x.astype("category"))
        return df

    def tolerance_bands(self, window: int | str, pct: float):
        df = self._obj
        rolling = df.rolling(window=window, center=True, min_periods=1)
        u_band = np.maximum(rolling.max(), df * (pct + 1))
        l_band = np.minimum(rolling.min(), df * (1 - pct))
        result = pd.concat([df, l_band, u_band], axis=1)
        cols = [f"{x}_{y}" for x in ["real", "lower", "upper"] for y in df.columns]
        result.set_axis(cols, axis="columns")
        return result

    def merge_columns(self, columns: list[Hashable], divider: str = " "):
        def split(x):
            return divider.join(str(i) for i in x)

        return self._obj[columns].apply(split, axis=1)

    def duplicate_features(self, columns: list[Hashable]):
        df = self._obj
        new_names = helpers.uniquify_names(columns, df.columns)
        new_cols = df[columns]
        new_cols.columns = new_names
        df = pd.concat([df, new_cols], axis=1)
        return df

    def eval(self, code: str, variable_name: str = "df"):
        """Apply a script to the dataset."""
        context = {variable_name: self._obj, "__builtins__": __builtins__}
        df = helpers.evaluate(code=code, context=context, return_val=variable_name)
        if not isinstance(df, pd.DataFrame):
            raise TypeError("Function needs to return a dataframe")
        return df

    @classmethod
    def from_script(cls, code: str, variable_name: str = "df"):
        """Return a ds resulting from a code block.

        Result is wrapped as function
        because we dont want "ds" hardcoded
        """
        context = {"__builtins__": __builtins__}
        df = helpers.evaluate(code, context, return_val=variable_name)
        if not isinstance(df, pd.DataFrame):
            raise TypeError("Function needs to return a dataframe")
        return df


if __name__ == "__main__":
    test = pd.DataFrame()
    test.pt.uniquify_columns()



import ast
import itertools
import logging
import time
from typing import Any, Optional


logger = logging.getLogger(__name__)


def uniquify_names(new_cols, old_cols):
    cols = []
    for col in new_cols:
        if col not in old_cols:
            cols.append(col)
            continue
        for i in itertools.count():
            name = f"{col}_{i}"
            if name not in old_cols:
                cols.append(name)
                break
    return cols


def evaluate(
    code, context: Optional[dict[str, Any]] = None, return_val: Optional[str] = None
):
    """Evaluate several lines of input, returning the result of the last line."""
    now = time.time()
    if context is None:
        context = locals()
    code = str(code) if return_val is None else f"{code}\n{return_val}"
    logger.debug("Evaluating code:\n%s", code)
    tree = ast.parse(code)
    eval_expr = ast.Expression(tree.body[-1].value)  # type: ignore
    # exec_expr = ast.Module(tree.body[:-1])  # type: ignore
    exec_expr = ast.parse("")
    exec_expr.body = tree.body[:-1]
    compiled = compile(exec_expr, "file", "exec")
    exec(compiled, context)
    val = eval(compile(eval_expr, "file", "eval"), context)
    logger.debug("Code evaluation took %s seconds.", time.time() - now)
    return val



from typing import Any

from bidict import bidict
import numpy as np
import pandas as pd


class DType:
    name: str
    pandas_str: str
    dtype: Any
    is_index_dtype = True

    @classmethod
    def get_dtypes(cls) -> bidict[str, str]:
        dct = {i.pandas_str: i.name for i in cls.__subclasses__()}
        return bidict(dct)

    @classmethod
    def get_dtypes_for_index(cls) -> bidict[str, str]:
        dct = {i.pandas_str: i.name for i in cls.__subclasses__() if i.is_index_dtype}
        return bidict(dct)


class Int8(DType):
    name = "Integer (8 bit)"
    pandas_str = "Int8"
    dtype = pd.Int8Dtype()
    # np_dtype = np.int8


class Int16(DType):
    name = "Integer (16 bit)"
    pandas_str = "Int16"
    dtype = pd.Int16Dtype()
    # np_dtype = np.int16


class Int64(DType):
    name = "Integer (64 bit)"
    pandas_str = "Int64"
    dtype = pd.Int64Dtype()
    # np_dtype = np.int64


class Float32(DType):
    name = "Float (32 bit)"
    pandas_str = "float32"
    dtype = np.float32
    # np_dtype = np.float32


class Float64(DType):
    name = "Float (64 bit)"
    pandas_str = "float64"
    dtype = np.float64
    # np_dtype = np.float64


class Category(DType):
    name = "Category"
    pandas_str = "category"
    dtype = pd.CategoricalDtype()


class DateTime(DType):
    name = "DateTime"
    pandas_str = "datetime64[ns, UTC]"
    dtype = pd.DatetimeTZDtype(tz="UTC")


class String(DType):
    name = "String"
    pandas_str = "string"
    dtype = pd.StringDtype()


class Bool(DType):
    name = "Bool"
    pandas_str = "bool"
    dtype = bool
    is_index_dtype = False


class Boolean(DType):
    name = "Boolean"
    pandas_str = "boolean"
    dtype = pd.BooleanDtype()
    is_index_dtype = False


class Object(DType):
    name = "Object"
    pandas_str = "object"
    dtype = object
    is_index_dtype = False



"""contains accessors for pandas."""

from .dataframe import DataFrameAccessor
from .index import IndexAccessor
from .series import SeriesAccessor

__all__ = ["DataFrameAccessor", "SeriesAccessor", "IndexAccessor", "scipydataframe"]



from collections.abc import Hashable
import logging
import time

# import numba
import numpy as np
import pandas as pd


logger = logging.getLogger(__name__)


def add_transition_info(
    ds: pd.DataFrame,
    colname: Hashable,
    threshold: float,
    extra_rows: int = 0,
    transition_col: str = "process_num",
    time_col: str = "secs",
) -> pd.DataFrame:
    """Calculate splits and add "secs" and "process_num" column to dataframe."""
    logger.info("Start splitting process")
    now = time.time()
    # setze ersten Wert auf 0. (evtl auch noch den letzten?)
    # Bisschen unschön, aber nötig, damit start und end indexes in sync sind.
    col = ds[colname]
    start_indexes = get_transition_indices(col.to_numpy(), threshold)
    start_indexes = np.maximum(start_indexes - extra_rows, 0)
    end_indexes = get_transition_indices(col.to_numpy(), threshold, True)
    end_indexes = np.minimum(end_indexes + extra_rows, len(ds.index))
    if len(start_indexes) > len(end_indexes):
        start_indexes = start_indexes[:-1]
    if len(end_indexes) > len(start_indexes):
        end_indexes = start_indexes[1:]
    if not colname or len(start_indexes) == 0 or len(end_indexes) == 0:
        start_indexes = [0]
        end_indexes = [len(ds.index)]  # perhaps -1 ?

    logger.info("Found indices. Applying new columns to dataset....")
    ds = ds.drop(time_col, errors="ignore", axis=1)
    proc = np.full((len(ds.index),), np.nan)
    for i, (start, end) in enumerate(zip(start_indexes, end_indexes), start=1):
        proc[start:end] = i
    categories = [i + 1 for i in range(len(start_indexes))]
    ds[transition_col] = pd.Categorical(proc, categories=categories)
    ds[time_col] = ds.groupby(transition_col).apply(calc_secs)[time_col]
    logger.info(f"Splitting took {(time.time() - now):.2f} seconds")
    return ds


def calc_secs(ds: pd.DataFrame) -> pd.DataFrame:
    idx = ds.index.astype(int).to_numpy() / 1_000_000_000
    ds["secs"] = idx - idx[0] if len(ds.index) > 0 else np.nan
    return ds


# @numba.jit(nopython=True, parallel=True)
def get_transition_indices(
    y: np.ndarray, threshold: float, falling_edge: bool = False
) -> np.ndarray:
    """Return indices where a transition occurs (default: detect rising edges)."""
    # Find where y crosses a threshold in a specific direction.
    lower = y < threshold
    higher = y >= threshold
    if falling_edge:
        return np.where(higher[:-1] & lower[1:])[0]
    else:
        return np.where(lower[:-1] & higher[1:])[0]


# @numba.jit(nopython=True, parallel=True)
def find_transition_times(
    t: np.ndarray, y: np.ndarray, threshold: float, falling_edge: bool = False
) -> np.ndarray:
    """Find transition times.

    Given the input signal `y` with samples at times `t`,
    find the times where `y` increases through the value `threshold`.

    `t` and `y` must be 1-D numpy arrays.

    Linear interpolation is used to estimate the time `t` between
    samples at which the transitions occur.
    """
    transition_indices = get_transition_indices(y, threshold, falling_edge)

    t0 = t[transition_indices]
    t1 = t[transition_indices + 1]
    y0 = y[transition_indices]
    y1 = y[transition_indices + 1]
    slope = (y1 - y0) / (t1 - t0)
    transition_times = t0 + (threshold - y0) / slope

    return transition_times


# @numba.jit(nopython=True)
def periods(t: np.ndarray, y: np.ndarray, threshold: float):
    """Return periods where y crosses the threshold over given time.

    Given the input signal `y` with samples at times `t`,
    find the time periods between the times at which the
    signal `y` increases through the value `threshold`.

    `t` and `y` must be 1-D numpy arrays.
    periods(df.index.astype(int).to_numpy(), df.InductorDown.to_numpy(), 0.5)
    """
    transition_times = find_transition_times(t, y, threshold)
    deltas = np.diff(transition_times)
    return deltas


def sliced(df: pd.DataFrame, column_name: str, threshold: float):
    t_indices = get_transition_indices(df[column_name], threshold)
    return np.split(df, t_indices)



import pandas as pd
import scipy.fftpack
import scipy.io
import scipy.io.arff


@pd.api.extensions.register_dataframe_accessor("pt_scipy")
class DataFrameAccessor:
    def __init__(self, parent):
        self._obj = parent

    def fft(self):
        ds = self._obj
        ds.index = ds.index.pt.index_to_secs()
        sample_rate = ds.index[1] - ds.index[0]
        # TODO: transform seems to give different results than apply, dunno why atm...
        ds = ds.transform(lambda x: scipy.fftpack.rfft(x.to_numpy(), len(ds.index)))
        ds.index = scipy.fftpack.rfftfreq(len(ds.index), sample_rate)
        ds = ds.abs()
        return ds

        # @staticmethod
        # def load_matlab(path):
        #     mat = scipy.io.loadmat(str(path))
        # mdata = mat["measuredData"]
        # ndata = {n: mdata[n][0, 0] for n in mdata.dtype.names}
        # cols = [n for n, v in ndata.items() if v.size == ndata["numIntervals"]]
        # ds = pd.DataFrame(data=np.concatenate([ndata[c] for c in cols], axis=1),
        #                   index=[datetime.datetime(*ts) for ts in ndata["timestamps"]],
        #                   columns=cols)
        # ds = pd.DataFrame(np.hstack((mat['X'], mat['y'])))
        # return ds

    @staticmethod
    def load_arff(path):
        data, meta = scipy.io.arff.loadarff(str(path))
        ds = pd.DataFrame(data=data, columns=meta.names())
        for i, t in enumerate(meta.types()):
            if t == "nominal":
                ds.iloc[:, i] = ds.iloc[:, i].str.decode("utf-8").astype("category")
        return ds


if __name__ == "__main__":
    import pydataset

    import pandastools  # type: ignore  # noqa: F401

    test = pydataset.data("Formaldehyde")
    result = test.pt_scipy.fft()
    print(result)



from __future__ import annotations

import logging
from typing import Literal

import pandas as pd


logger = logging.getLogger(__name__)


@pd.api.extensions.register_index_accessor("pt")
class IndexAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def index_to_secs(self):
        if self._obj.size == 0:
            logger.debug("index_to_secs failed. Index empty")
            return self._obj
        elif isinstance(self._obj, pd.DatetimeIndex):
            secs = self._obj.view(int) / 1_000_000_000
            secs = secs - secs[0]
            return secs
        else:
            logger.debug("index_to_secs failed. No DateTimeIndex")
            return self._obj

    def detect_gaps(self, max_diff: float | str, level=None):
        if level is None:
            index = self._obj
        else:
            index = self._obj.get_level_values(level)
        return (index.to_series().diff() > max_diff).cumsum() + 1

    def to_datetime(
        self,
        level=None,
        fmt: str | None = None,
        errors: Literal["ignore", "raise"] = "ignore",
    ):
        if isinstance(self._obj, pd.MultiIndex):
            values = self._obj.get_level_values(level).to_series()
            values = pd.to_datetime(
                values, format=fmt, errors=errors, infer_datetime_format=not bool(fmt)
            )
            return self._obj.set_levels(values, level)
        else:
            return pd.to_datetime(
                self._obj, format=fmt, errors=errors, infer_datetime_format=not bool(fmt)
            )


if __name__ == "__main__":
    df = pd.DataFrame(dict(a=[1, 2, 3, 4, 5]))
    print(df["a"].pt.tolerance_bands(5, 0.01))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pandastools', 'pandastools.accessors', 'pandastools.utils']

package_data = \
{'': ['*']}

install_requires = \
['bidict>=0,<1', 'pandas>=1.4,<2.0']

extras_require = \
{'addons:python_version >= "3.9" and python_version < "3.12"': ['scipy>=1.7,<2.0']}

setup_kwargs = {
    'name': 'pandastools',
    'version': '0.6.1',
    'description': 'Helper functions for Pandas DataFrames / Series',
    'long_description': '# pandastools: Helper functions for Pandas DataFrames / Series\n[![PyPI Latest Release](https://img.shields.io/pypi/v/pandastools.svg)](https://pypi.org/project/pandastools/)\n[![Package Status](https://img.shields.io/pypi/status/pandastools.svg)](https://pypi.org/project/pandastools/)\n[![License](https://img.shields.io/pypi/l/pandastools.svg)](https://github.com/pandastools-dev/pandastools/blob/master/LICENSE)\n[![Travis Build Status](https://travis-ci.org/pandastools-dev/pandastools.svg?branch=master)](https://travis-ci.org/pandastools-dev/pandastools)\n[![CodeCov](https://codecov.io/gh/phil65/PandasTools/branch/master/graph/badge.svg)](https://codecov.io/gh/phil65/PandasTools)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![PyUp](https://pyup.io/repos/github/phil65/PandasTools/shield.svg)](https://pyup.io/repos/github/phil65/PandasTools/)\n\n## What is it?\n\n**PandasTools** is a Python package that provides additional functions for Pandas DataFrames, Series and Indexes via accessors\n\n## Main Features\n\n  - TODO\n\n   [dataframe]: https://phil65.github.io/PandasTools/dataframe.html\n\n\n## Where to get it\nThe source code is currently hosted on GitHub at:\nhttps://github.com/phil65/PandasTools\n\nThe latest released version are available at the [Python\npackage index](https://pypi.org/project/pandastools).\n\n```sh\n# or PyPI\npip install pandastools\n```\n\n## Dependencies\n- [pandas](https://pypi.org/project/pandas)\n\n\n## Installation from sources\n\nThis project uses poetry for dependency management and packaging. Install this first.\nIn the `pandastools` directory (same one where you found this file after\ncloning the git repo), execute:\n\n```sh\npoetry install\n```\n\n## License\n[MIT](LICENSE)\n\n## Documentation\nThe official documentation is hosted on Github Pages: https://phil65.github.io/PandasTools/\n\n## Contributing to pandas [![Open Source Helpers](https://www.codetriage.com/phil65/pandastools/badges/users.svg)](https://www.codetriage.com/phil65/pandastools)\n\nAll contributions, bug reports, bug fixes, documentation improvements, enhancements, and ideas are welcome.\n\nOr maybe through using PandasTools you have an idea of your own or are looking for something in the documentation and thinking ‘this can be improved’...you can do something about it!\n',
    'author': 'phil65',
    'author_email': 'philipptemminghoff@googlemail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/phil65/pandastools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
